"""
Array (Tensor) functions
"""

from __future__ import annotations
from typing import Union, TypeVar
from returnn.tensor import Tensor
from ._backend import global_backend
from .types import RawTensorTypes

T = TypeVar("T")


def convert_to_tensor(value: Union[Tensor, T, RawTensorTypes]) -> Tensor[T]:
    """
    :param value: tensor, or scalar raw tensor or some other scalar value
    :return: tensor
    """
    if isinstance(value, Tensor):  # fast path
        return value
    return global_backend.convert_to_tensor(value=value)



"""
Uses KenLM (https://kheafield.com/code/kenlm/) (extern/kenlm) to read n-gram LMs (ARPA format),
and provides a TF op to use them.

"""

import sys
import os
import tensorflow as tf

returnn_dir = os.path.dirname(os.path.abspath(__file__))
kenlm_dir = returnn_dir + "/extern/kenlm"


def kenlm_checked_out():
    """
    :rtype: bool
    """
    return os.path.exists("%s/lm/test.arpa" % kenlm_dir)


# https://www.tensorflow.org/guide/extend/op
# Also see TFUtil.TFArrayContainer for TF resources.
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/tensor.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/tensor_shape.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/tensor_types.h
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/lib/strings/str_util.h
_src_code = """
#include <exception>
#include "tensorflow/core/framework/op.h"
#include "tensorflow/core/framework/op_kernel.h"
#include "tensorflow/core/framework/shape_inference.h"
#include "tensorflow/core/framework/resource_mgr.h"
#include "tensorflow/core/framework/resource_op_kernel.h"
#include "tensorflow/core/framework/tensor.h"
#include "tensorflow/core/platform/macros.h"
#include "tensorflow/core/platform/mutex.h"
#include "tensorflow/core/platform/types.h"
#include "tensorflow/core/public/version.h"


using namespace tensorflow;

#if (TF_MAJOR_VERSION < 2) || (TF_MAJOR_VERSION == 2 && TF_MINOR_VERSION < 2)
using tstring = std::string;
#else
using tstring = tensorflow::tstring;
#endif


REGISTER_OP("KenLmLoadModel")
.Attr("filename: string")
.Attr("container: string = ''")
.Attr("shared_name: string = ''")
.Output("handle: resource")
.SetIsStateful()
.SetShapeFn(shape_inference::ScalarShape)
.Doc("KenLmLoadModel: loads KenLM model, creates TF resource, persistent across runs in the session");


REGISTER_OP("KenLmAbsScoreStrings")
.Input("handle: resource")
.Input("strings: string")
.Output("scores: float32")
.SetShapeFn([](::tensorflow::shape_inference::InferenceContext* c) {
  c->set_output(0, c->input(1));
  return Status::OK();
})
.Doc("KenLmScoreStrings: scores texts. returns in +log space (natural log, not base 10)");


REGISTER_OP("KenLmAbsScoreBpeStrings")
.Input("handle: resource")
.Input("bpe_merge_symbol: string")
.Input("strings: string")
.Output("scores: float32")
.SetShapeFn([](::tensorflow::shape_inference::InferenceContext* c) {
  c->set_output(0, c->input(2));
  return Status::OK();
})
.Doc("KenLmAbsScoreBpeStrings: optionally BPE-merges, remove surrounding whitespaces and scores texts."
  " returns in +log space (natural log, not base 10)."
  " relative score, relative to previous text."
  );


REGISTER_OP("KenLmAbsScoreBpeStringsDense")
.Input("handle: resource")
.Input("bpe_merge_symbol: string")
.Input("strings: string")
.Input("labels: string")
.Output("scores: float32")
.Output("dense_scores: float32")
.SetShapeFn([](::tensorflow::shape_inference::InferenceContext* c) {
  c->set_output(0, c->input(2));
  ::tensorflow::shape_inference::ShapeHandle out_shape;
  TF_RETURN_IF_ERROR(c->Concatenate(c->input(2), c->input(3), &out_shape));
  c->set_output(1, out_shape);
  return Status::OK();
})
.Doc("KenLmAbsScoreBpeStrings: optionally BPE-merges, remove surrounding whitespaces and scores texts."
  " returns in +log space (natural log, not base 10)."
  " relative score, relative to previous text."
  " dense output, for all possible succeeding labels.");


// https://github.com/kpu/kenlm/blob/master/lm/model.hh
// https://github.com/kpu/kenlm/blob/master/lm/virtual_interface.hh
// https://github.com/kpu/kenlm/blob/master/python/kenlm.pyx
struct KenLmModel : public ResourceBase {
  explicit KenLmModel(const string& filename)
      : filename_(filename), model_(filename.c_str()) {}

  float abs_score(const ::tstring& text) {
    float total = 0;
    mutex_lock l(mu_);
    lm::ngram::State state, out_state;
    model_.BeginSentenceWrite(&state);
    for(const string& word : tensorflow::str_util::Split(text, ' ')) {
      if(word.empty()) continue;
      auto word_idx = model_.BaseVocabulary().Index(word);
      total += model_.FullScore(state, word_idx, out_state).prob;
      state = out_state;
    }
    // KenLM returns score in +log10 space.
    // We want to return in (natural) +log space.
    // 10 ** x = e ** (x * log(10))
    return total * logf(10.);
  }

  // See comments below.
  // We expect that the text either ends with a space or not, i.e. "... word " or "... subword".
  float abs_score_dense(
        const ::tstring& text, const ::tstring& last_word_join,
        const TTypes<::tstring>::ConstFlat labels, TTypes<float>::UnalignedFlat out_dense_scores) {
    assert(labels.size() == out_dense_scores.size());
    mutex_lock l(mu_);
    lm::ngram::State state, out_state;
    model_.BeginSentenceWrite(&state);
    // We expect that the text either ends with a space or not, i.e. "... word " or "... subword".
    // We split the text into words. In the first case, we would have an empty word at the end, otherwise not.
    auto words = tensorflow::str_util::Split(text, ' ');
    float total_score = 0;
    ::tstring last_word = "";
    if(!words.empty()) {
      last_word = words[words.size() - 1];
      // Only up to the last word, which is either empty or a subword, which we join below.
      for(int i = 0; i < words.size() - 1; ++i) {
        auto word = words[i];
        if(word.empty()) continue;
        auto word_idx = model_.BaseVocabulary().Index(word);
        total_score += model_.FullScore(state, word_idx, out_state).prob;
        state = out_state;
      }
    }
    for(int i = 0; i < labels.size(); ++i) {
      ::tstring word = last_word + labels(i);
      auto word_idx = model_.BaseVocabulary().Index(word);
      float score = model_.FullScore(state, word_idx, out_state).prob;
      out_dense_scores(i) = (total_score + score) * logf(10.);
    }
    // Return the score from the prev step.
    if(!last_word.empty()) {
      auto word_idx = model_.BaseVocabulary().Index(last_word + last_word_join);
      total_score += model_.FullScore(state, word_idx, out_state).prob;
    }
    return total_score * logf(10.);
  }

  string DebugString()
#if (TF_MAJOR_VERSION == 1 && TF_MINOR_VERSION >= 14) || (TF_MAJOR_VERSION > 1)
const
#endif
  override {
    return strings::StrCat("KenLmModel[", filename_, "]");
  }

  const string filename_;
  mutex mu_;
  lm::ngram::ProbingModel model_;
};


// https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/resource_op_kernel.h
// TFUtil.TFArrayContainer
class KenLmLoadModelOp : public ResourceOpKernel<KenLmModel> {
 public:
  explicit KenLmLoadModelOp(OpKernelConstruction* context)
      : ResourceOpKernel(context) {
    OP_REQUIRES_OK(context, context->GetAttr("filename", &filename_));
  }

 private:
  virtual bool IsCancellable() const { return false; }
  virtual void Cancel() {}

  Status CreateResource(KenLmModel** ret) override {
    try {
      *ret = new KenLmModel(filename_);
    } catch (std::exception& exc) {
      return errors::Internal("Could not load KenLmModel ", filename_, ", exception: ", exc.what());
    }
    if(*ret == nullptr)
      return errors::ResourceExhausted("Failed to allocate");
    return Status::OK();
  }

  Status VerifyResource(KenLmModel* lm) override {
    if(lm->filename_ != filename_)
      return errors::InvalidArgument("Filename mismatch: expected ", filename_,
                                     " but got ", lm->filename_, ".");
    return Status::OK();
  }

  string filename_;
};

REGISTER_KERNEL_BUILDER(Name("KenLmLoadModel").Device(DEVICE_CPU), KenLmLoadModelOp);


class KenLmAbsScoreStringsOp : public OpKernel {
 public:
  using OpKernel::OpKernel;

  void Compute(OpKernelContext* context) override {
    KenLmModel* lm;
    {
      const Tensor* handle;
      OP_REQUIRES_OK(context, context->input("handle", &handle));
      OP_REQUIRES_OK(context, GetResourceFromContext(context, "handle", &lm));
    }
    core::ScopedUnref unref(lm);

    const Tensor& input_tensor = context->input(1);
    auto input_flat = input_tensor.flat<::tstring>();

    Tensor* output_tensor = NULL;
    OP_REQUIRES_OK(context, context->allocate_output(0, input_tensor.shape(), &output_tensor));
    auto output_flat = output_tensor->flat<float>();

    for(int i = 0; i < input_flat.size(); ++i) {
      output_flat(i) = lm->abs_score(input_flat(i));
    }
  }
};

REGISTER_KERNEL_BUILDER(Name("KenLmAbsScoreStrings").Device(DEVICE_CPU), KenLmAbsScoreStringsOp);


class KenLmAbsScoreBpeStringsOp : public OpKernel {
 public:
  using OpKernel::OpKernel;

  void Compute(OpKernelContext* context) override {
    KenLmModel* lm;
    {
      const Tensor* handle;
      OP_REQUIRES_OK(context, context->input("handle", &handle));
      OP_REQUIRES_OK(context, GetResourceFromContext(context, "handle", &lm));
    }
    core::ScopedUnref unref(lm);

    OP_REQUIRES(context, context->input(1).NumElements() == 1,
      errors::InvalidArgument(
        "bpe_merge_symbol must be a single element but got shape ",
        context->input(1).shape().DebugString()));
    const auto& bpe_merge_symbol = context->input(1).flat<::tstring>()(0);

    const Tensor& input_tensor = context->input(2);
    auto input_flat = input_tensor.flat<::tstring>();

    Tensor* output_tensor = NULL;
    OP_REQUIRES_OK(context, context->allocate_output(0, input_tensor.shape(), &output_tensor));
    auto output_flat = output_tensor->flat<float>();

    for(int i = 0; i < input_flat.size(); ++i) {
      ::tstring text = input_flat(i);
      if(!bpe_merge_symbol.empty())
        text = tensorflow::str_util::StringReplace(text, bpe_merge_symbol + " ", "", /* replace_all */ true);
      tensorflow::StringPiece sp(text);
      tensorflow::str_util::RemoveWhitespaceContext(&sp);
      text = std::string(sp.data(), sp.size());
      output_flat(i) = lm->abs_score(text);
    }
  }
};

REGISTER_KERNEL_BUILDER(Name("KenLmAbsScoreBpeStrings").Device(DEVICE_CPU), KenLmAbsScoreBpeStringsOp);


class KenLmAbsScoreBpeStringsDenseOp : public OpKernel {
 public:
  using OpKernel::OpKernel;

  void Compute(OpKernelContext* context) override {
    KenLmModel* lm;
    {
      const Tensor* handle;
      OP_REQUIRES_OK(context, context->input("handle", &handle));
      OP_REQUIRES_OK(context, GetResourceFromContext(context, "handle", &lm));
    }
    core::ScopedUnref unref(lm);

    OP_REQUIRES(context, context->input(1).NumElements() == 1,
      errors::InvalidArgument(
        "bpe_merge_symbol must be a single element but got shape ",
        context->input(1).shape().DebugString()));
    const ::tstring& bpe_merge_symbol = context->input(1).flat<::tstring>()(0);

    const Tensor& input_tensor = context->input(2);
    auto input_flat = input_tensor.flat<::tstring>();

    const Tensor& labels_tensor = context->input(3);
    auto labels_flat = labels_tensor.flat<::tstring>();

    Tensor* output_tensor = NULL;
    OP_REQUIRES_OK(context, context->allocate_output(0, input_tensor.shape(), &output_tensor));
    auto output_flat = output_tensor->flat<float>();

    Tensor* output_dense_tensor = NULL;
    TensorShape output_dense_shape(input_tensor.shape());
    output_dense_shape.AppendShape(labels_tensor.shape());
    OP_REQUIRES_OK(context, context->allocate_output(1, output_dense_shape, &output_dense_tensor));
    Tensor output_dense_flat_tensor;
    OP_REQUIRES(context,
      output_dense_flat_tensor.CopyFrom(
        *output_dense_tensor,
        TensorShape({input_tensor.NumElements(), labels_tensor.NumElements()})),
      errors::Internal("CopyFrom failed"));

    for(int i = 0; i < input_flat.size(); ++i) {
      ::tstring text = input_flat(i);
      if(!bpe_merge_symbol.empty())
        text = tensorflow::str_util::StringReplace(text, bpe_merge_symbol + " ", "", /* replace_all */ true);
      output_flat(i) = lm->abs_score_dense(
        text, bpe_merge_symbol, labels_flat, output_dense_flat_tensor.Slice(i, i + 1).unaligned_flat<float>());
    }
  }
};

REGISTER_KERNEL_BUILDER(Name("KenLmAbsScoreBpeStringsDense").Device(DEVICE_CPU), KenLmAbsScoreBpeStringsDenseOp);

"""

_kenlm_src_code_workarounds = """
// ------- start with some workarounds { ------
// The KenLM code (util/integer_to_string.cc) includes this file in the wrong namespace.
// Thus include it here now.
#include <emmintrin.h>
// ------- end with workarounds } -------------
"""


_tf_mod = None


def get_tf_mod(verbose=False):
    """
    :param bool verbose:
    :return: module
    """
    global _tf_mod
    if _tf_mod:
        return _tf_mod
    import platform
    from glob import glob
    from returnn.tf.util.basic import OpCodeCompiler

    # References:
    # https://github.com/kpu/kenlm/blob/master/setup.py
    # https://github.com/kpu/kenlm/blob/master/compile_query_only.sh

    # Collect files.
    assert kenlm_checked_out(), "submodule in %r not checked out?" % kenlm_dir
    files = glob("%s/util/*.cc" % kenlm_dir)
    files += glob("%s/lm/*.cc" % kenlm_dir)
    files += glob("%s/util/double-conversion/*.cc" % kenlm_dir)
    files = [fn for fn in files if not (fn.endswith("main.cc") or fn.endswith("test.cc"))]
    assert files, "submodule in %r not checked out?" % kenlm_dir
    libs = ["z"]
    if platform.system() != "Darwin":
        libs.append("rt")

    # Put code all together in one big blob.
    src_code = ""
    src_code += _kenlm_src_code_workarounds
    for fn in files:
        f_code = open(fn).read()
        f_code = "".join([x for x in f_code if ord(x) < 128])  # enforce ASCII
        # We need to do some replacements to not clash symbol names.
        fn_short = os.path.basename(fn).replace(".", "_")
        for word in ["kConverter"]:
            f_code = f_code.replace(word, "%s_%s" % (fn_short, word))
        src_code += "\n// ------------ %s : BEGIN { ------------\n" % os.path.basename(fn)
        # https://gcc.gnu.org/onlinedocs/cpp/Line-Control.html#Line-Control
        src_code += '#line 1 "%s"\n' % os.path.basename(fn)
        src_code += f_code
        src_code += "\n// ------------ %s : END } --------------\n\n" % os.path.basename(fn)
    src_code += "\n\n// ------------ our code now: ------------\n\n"
    src_code += '#line 1 "<our code>"\n'
    src_code += _src_code

    compiler = OpCodeCompiler(
        base_name="KenLM",
        code_version=1,
        code=src_code,
        include_paths=(kenlm_dir, kenlm_dir + "/util/double-conversion"),
        c_macro_defines={"NDEBUG": 1, "KENLM_MAX_ORDER": 6, "HAVE_ZLIB": 1},
        ld_flags=["-l%s" % lib for lib in libs],
        is_cpp=True,
        use_cuda_if_available=False,
        verbose=verbose,
    )
    tf_mod = compiler.load_tf_module()
    assert hasattr(tf_mod, "ken_lm_abs_score_strings"), "content of mod: %r" % (dir(tf_mod),)
    _tf_mod = tf_mod
    return tf_mod


def ken_lm_load(filename):
    """
    :param str filename:
    :return: TF resource handle
    :rtype: tf.Tensor
    """
    return get_tf_mod().ken_lm_load_model(filename=filename)


def ken_lm_abs_score_strings(handle, strings):
    """
    :param tf.Tensor handle: TF resource handle returned by :func:`ken_lm_load`
    :param tf.Tensor strings: strings which are being scores. white-space delimited words.
    :return: same shape as `strings`, float32
    :rtype: tf.Tensor
    """
    return get_tf_mod().ken_lm_abs_score_strings(handle=handle, strings=strings)


def ken_lm_abs_score_bpe_strings(handle, bpe_merge_symbol, strings):
    """
    :param tf.Tensor handle: TF resource handle returned by :func:`ken_lm_load`
    :param str bpe_merge_symbol: e.g. "@@"
    :param tf.Tensor strings: strings which are being scores. white-space delimited words.
    :return: same shape as `strings`, float32
    :rtype: tf.Tensor
    """
    return get_tf_mod().ken_lm_abs_score_bpe_strings(handle=handle, bpe_merge_symbol=bpe_merge_symbol, strings=strings)


def ken_lm_abs_score_bpe_strings_dense(handle, bpe_merge_symbol, strings, labels):
    """
    :param tf.Tensor handle: TF resource handle returned by :func:`ken_lm_load`
    :param str bpe_merge_symbol: e.g. "@@"
    :param tf.Tensor strings: strings which are being scores. white-space delimited words.
    :param tf.Tensor|tf.Variable labels:
    :return: same shape as `strings`, float32
    :rtype: tf.Tensor
    """
    return get_tf_mod().ken_lm_abs_score_bpe_strings_dense(
        handle=handle, bpe_merge_symbol=bpe_merge_symbol, strings=strings, labels=labels
    )


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    # Try to compile now.
    get_tf_mod(verbose=True)
    # Some demo.
    input_strings = sys.argv[1:] or ["hello world </s>"]
    test_lm_file = kenlm_dir + "/lm/test.arpa"
    assert os.path.exists(test_lm_file)
    lm_tf = ken_lm_load(filename=test_lm_file)
    input_strings_tf = tf.compat.v1.placeholder(tf.string, [None])
    output_scores_tf = ken_lm_abs_score_strings(handle=lm_tf, strings=input_strings_tf)
    with tf.compat.v1.Session() as session:
        output_scores = session.run(output_scores_tf, feed_dict={input_strings_tf: input_strings})
        print("input strings:", input_strings, "(sys.argv[1:])")
        print("output scores:", output_scores)



"""
This package provides the frontend API.
The frontend API supports multiple backends.
https://github.com/rwth-i6/returnn/issues/1120

The convention for the user is to do::

    from returnn import frontend as rf
"""

# Not necessarily all here, but the most common ones.
# (Take PyTorch `torch.nn` as a reference.)
from .array_ import *
from .dims import *
from .dot import *
from .outputs import *
from .reduce import *
from .types import *



"""
Reduce
"""

from __future__ import annotations
from typing import Union, TypeVar, Sequence
from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim
from ._backend import get_backend_by_tensor

T = TypeVar("T")


def reduce(
    source: Tensor[T],
    *,
    mode: str,
    axis: Union[Dim, Sequence[Dim]],
    use_time_mask: bool = NotSpecified,
) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param mode: "sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"
    :param axis:
    :param use_time_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    rf = get_backend_by_tensor(source)
    return rf.reduce(source=source, mode=mode, axis=axis, use_time_mask=use_time_mask)



"""
Backend for exposing PyTorch-specific functionality.
"""

from __future__ import annotations
from typing import Optional, Union, Sequence, Tuple

import torch
from returnn.tensor import Tensor, Dim
from returnn.util.basic import prod

# noinspection PyProtectedMember
from returnn.frontend._backend import Backend
from returnn.frontend import RawTensorTypes


_TT = Tensor[torch.Tensor]


# Ignore this warning until we really expect that we implemented everything.
# noinspection PyAbstractClass
class TorchBackend(Backend[torch.Tensor]):
    """
    PyTorch backend
    """

    RawTensorType = torch.Tensor

    @staticmethod
    def get_dtype_name_raw(raw_tensor: torch.Tensor) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        return str(raw_tensor.dtype).replace("torch.", "")

    @staticmethod
    def get_ndim_raw(raw_tensor: torch.Tensor) -> int:
        """
        :return: ndim of raw tensor
        """
        return raw_tensor.dim()

    @staticmethod
    def get_known_shape_raw(raw_tensor: torch.Tensor) -> Tuple[Optional[int]]:
        """
        :return: shape of raw tensor; here for PyTorch the full shape is always known
        """
        return tuple(raw_tensor.size())

    @staticmethod
    def expand_dims_raw(raw_tensor: torch.Tensor, axis: int) -> torch.Tensor:
        """
        :param raw_tensor:
        :param axis: e.g. 1
        :return: raw tensor with new axis
        """
        return raw_tensor.unsqueeze(axis)

    @staticmethod
    def compare_raw(a: torch.Tensor, kind: str, b: torch.Tensor) -> torch.Tensor:
        """
        :param a:
        :param kind: "equal"|"==", "less"|"<", "less_equal"|"<=", "greater"|">", "greater_equal"|">=",
            "not_equal"|"!="|"<>"
        :param b:
        :return: a `kind` b
        """
        assert a.dim() == b.dim()
        kind = {
            "==": "eq",
            "<=": "less_equal",
            "<": "less",
            ">=": "greater_equal",
            ">": "greater",
            "!=": "not_equal",
            "<>": "not_equal",
        }.get(kind, kind)
        op = getattr(torch, kind)  # e.g. torch.equal
        return op(a, b)

    @staticmethod
    def combine_raw(a: torch.Tensor, kind: str, b: torch.Tensor) -> torch.Tensor:
        """
        :param a:
        :param kind: "add"|"+", "sub"|"-", "mul"|"*", "truediv"|"/", "floordiv"|"//", "mod"|"%", "pow"|"**",
            "max"|"maximum", "min"|"minimum", "logical_and", "logical_or", "squared_difference"
        :param b:
        :return: a `kind` b
        """
        assert a.dim() == b.dim()
        kind = {
            "+": "add",
            "-": "sub",
            "*": "mul",
            "/": "true_divide",
            "truediv": "true_divide",
            "//": "floor_divide",
            "floordiv": "floor_divide",
            "%": "remainder",  # Python-like modulo, not C-like (torch.fmod)
            "mod": "remainder",
            "**": "pow",
            "max": "maximum",
            "min": "minimum",
        }.get(kind, kind)
        op = getattr(torch, kind)  # e.g. torch.add
        return op(a, b)

    @staticmethod
    def convert_to_tensor(value: Union[Tensor, torch.Tensor, RawTensorTypes]) -> Tensor[torch.Tensor]:
        """
        :param value:
        :return: tensor
        """
        if isinstance(value, Tensor):
            return value
        value = torch.tensor(value)
        assert isinstance(value, torch.Tensor)
        assert value.shape.as_list() == [], f"scalar expected, got {value}"
        return Tensor("const", raw_tensor=value, dims=[], dtype=value.dtype.base_dtype.name)

    @staticmethod
    def dot(a: _TT, b: _TT, *, reduce: Union[Dim, Sequence[Dim]]) -> _TT:
        """
        dot-product of a and b, see base class doc string
        """
        if isinstance(reduce, Dim):
            reduce = [reduce]

        a_dims = a.dims
        b_dims = b.dims

        assert all(
            dim in a_dims for dim in reduce
        ), f"'a' does not have the specified reduce dim(s) {reduce} (a dims: {a_dims})"
        assert all(
            dim in b_dims for dim in reduce
        ), f"'b' does not have the specified reduce dim(s) {reduce} (b dims: {b_dims})"

        reduce_dims_sorted_by_a = [dim for dim in a_dims if dim in reduce]

        a_reduce_axes = [a_dims.index(reduce_dim) for reduce_dim in reduce_dims_sorted_by_a]
        b_reduce_axes = [b_dims.index(reduce_dim) for reduce_dim in reduce_dims_sorted_by_a]

        a_unique_axes = [i for (i, dim) in enumerate(a_dims) if dim not in b_dims]
        b_unique_axes = [i for (i, dim) in enumerate(b_dims) if dim not in a_dims]

        common_dims = [dim for dim in a_dims if dim in b_dims and dim not in reduce]

        a_common_axes = [a_dims.index(common_dim) for common_dim in common_dims]
        b_common_axes = [b_dims.index(common_dim) for common_dim in common_dims]

        a_raw = a.raw_tensor
        b_raw = b.raw_tensor

        a_shape = a_raw.shape
        b_shape = b_raw.shape

        common_axes_shape = tuple(a_shape[i] for i in a_common_axes)
        b_common_axes_shape = tuple(b_shape[i] for i in b_common_axes)
        assert common_axes_shape == b_common_axes_shape, "Tensor shape for common Dims of a and b does not match."

        common_axes_total_dimension = prod(common_axes_shape)

        a_unique_axes_shape = tuple(a_shape[i] for i in a_unique_axes)
        b_unique_axes_shape = tuple(b_shape[i] for i in b_unique_axes)

        a_unique_axes_total_dimension = prod(a_unique_axes_shape)
        b_unique_axes_total_dimension = prod(b_unique_axes_shape)

        reduce_axes_shape = tuple(a_shape[i] for i in a_reduce_axes)
        b_reduce_axes_shape = tuple(b_shape[i] for i in b_reduce_axes)
        assert reduce_axes_shape == b_reduce_axes_shape, "Tensor shape for reduce Dims does not match between a and b."

        reduce_axes_total_dimension = prod(reduce_axes_shape)

        a_raw = torch.permute(a_raw, a_common_axes + a_unique_axes + a_reduce_axes)
        b_raw = torch.permute(b_raw, b_common_axes + b_reduce_axes + b_unique_axes)

        if common_axes_total_dimension == 1:  # standard matrix multiplication
            a_raw = torch.reshape(a_raw, (a_unique_axes_total_dimension, reduce_axes_total_dimension))
            b_raw = torch.reshape(b_raw, (reduce_axes_total_dimension, b_unique_axes_total_dimension))

            raw_result = torch.mm(a_raw, b_raw)

        else:  # batched matrix multiplication
            a_raw = torch.reshape(
                a_raw, (common_axes_total_dimension, a_unique_axes_total_dimension, reduce_axes_total_dimension)
            )
            b_raw = torch.reshape(
                b_raw, (common_axes_total_dimension, reduce_axes_total_dimension, b_unique_axes_total_dimension)
            )

            raw_result = torch.bmm(a_raw, b_raw)

        raw_result = torch.reshape(raw_result, common_axes_shape + a_unique_axes_shape + b_unique_axes_shape)

        a_unique_dims = [a_dims[i] for i in a_unique_axes]
        b_unique_dims = [b_dims[i] for i in b_unique_axes]
        result_dims = common_dims + a_unique_dims + b_unique_dims

        result_tensor = Tensor(name="dot", dims=result_dims, raw_tensor=raw_result)

        return result_tensor



"""
tests for returnn.torch.frontend
"""

import _setup_test_env  # noqa

import torch
import pytest

from returnn.tensor import Tensor, Dim
from returnn.torch.frontend import TorchBackend as rf


def test_dot_scalar_multiplication():
    a_raw = torch.tensor(2.0)
    b_raw = torch.tensor(3.0)

    a = Tensor(name="a", dims=[], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[])

    assert pytest.approx(result.raw_tensor) == 6.0


def test_dot_scalar_product():
    a_raw = torch.tensor([1.0, 2.0, 3.0])
    b_raw = torch.tensor([4.0, 5.0, 6.0])

    feature_dim = Dim(dimension=3)

    a = Tensor(name="a", dims=[feature_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[feature_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[feature_dim])

    assert pytest.approx(result.raw_tensor) == 32.0


def test_dot_outer_product():
    a_raw = torch.tensor([1.0, 2.0, 3.0])
    b_raw = torch.tensor([4.0, 5.0, 6.0])

    a_feature_dim = Dim(dimension=3)
    b_feature_dim = Dim(dimension=3)

    a = Tensor(name="a", dims=[a_feature_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[b_feature_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[])

    assert result.dims == (a_feature_dim, b_feature_dim)
    assert result.raw_tensor.shape == (3, 3)


def test_dot_matrix_vector_product():
    a_raw = torch.tensor([[1.0, 2.0, 3.0], [-1.0, -2.0, -3.0]])
    b_raw = torch.tensor([4.0, 5.0])

    a_feature_dim = Dim(dimension=3)
    reduce_dim = Dim(dimension=2)

    a = Tensor(name="a", dims=[reduce_dim, a_feature_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[reduce_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[reduce_dim])

    assert result.dims == (a_feature_dim,)
    assert result.raw_tensor.tolist() == pytest.approx([-1.0, -2.0, -3.0])


def test_dot_matrix_matrix_product():
    a_raw = torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
    b_raw = torch.tensor([[1.0, -2.0], [2.0, -4.0]])

    a_feature_dim = Dim(dimension=3)
    b_feature_dim = Dim(dimension=2)
    reduce_dim = Dim(dimension=2)

    a = Tensor(name="a", dims=[a_feature_dim, reduce_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[reduce_dim, b_feature_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[reduce_dim])

    assert result.dims == (a_feature_dim, b_feature_dim)
    assert torch.allclose(result.raw_tensor, torch.tensor([[5.0, -10.0], [11.0, -22.0], [17.0, -34.0]]))


def test_dot_scale_matrix():
    a_raw = torch.tensor([[1.0, 2.0, 3.0], [-1.0, -2.0, -3.0]])
    b_raw = torch.tensor(2.0)

    a_feature_dim1 = Dim(dimension=2)
    a_feature_dim2 = Dim(dimension=3)

    a = Tensor(name="a", dims=[a_feature_dim1, a_feature_dim2], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[])

    assert result.dims == (a_feature_dim1, a_feature_dim2)
    assert torch.allclose(result.raw_tensor, torch.tensor([[2.0, 4.0, 6.0], [-2.0, -4.0, -6.0]]))


def test_dot_batched_scalar_multiplication():
    a_raw = torch.tensor([1.0, 2.0, 3.0])
    b_raw = torch.tensor([4.0, 5.0, 6.0])

    batch_dim = Dim(dimension=3)

    a = Tensor(name="a", dims=[batch_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[batch_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[])

    assert result.dims == (batch_dim,)
    assert result.raw_tensor.tolist() == pytest.approx([4.0, 10.0, 18.0])


def test_dot_batched_scalar_product():
    a_raw = torch.tensor([[1.0, 2.0, 3.0], [-1.0, -2.0, -3.0]])
    b_raw = torch.tensor([[4.0, 5.0, 6.0], [4.0, 5.0, 6.0]])

    batch_dim = Dim(dimension=2)
    feature_dim = Dim(dimension=3)

    a = Tensor(name="a", dims=[batch_dim, feature_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[batch_dim, feature_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[feature_dim])

    assert result.dims == (batch_dim,)
    assert result.raw_tensor.tolist() == pytest.approx([32.0, -32.0])


def test_dot_batched_outer_product():
    a_raw = torch.tensor([[1.0, 2.0, 3.0], [-1.0, -2.0, -3.0]])
    b_raw = torch.tensor([[4.0, 5.0, 6.0], [4.0, 5.0, 6.0]])

    batch_dim = Dim(dimension=2)
    a_feature_dim = Dim(dimension=3)
    b_feature_dim = Dim(dimension=3)

    a = Tensor(name="a", dims=[batch_dim, a_feature_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[batch_dim, b_feature_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[])

    assert result.dims == (batch_dim, a_feature_dim, b_feature_dim)
    assert result.raw_tensor.shape == (2, 3, 3)


def test_dot_batched_matrix_vector_product():
    a_raw = torch.tensor([[[1.0, -1.0], [2.0, -2.0]], [[3.0, -3.0], [4.0, -4.0]], [[5.0, -5.0], [6.0, -6.0]]])
    b_raw = torch.tensor([[1.0, 2.0], [2.0, 4.0]])

    batch_dim = Dim(dimension=2)
    a_feature_dim = Dim(dimension=3)
    reduce_dim = Dim(dimension=2)

    # intentionally test strange batch axis
    a = Tensor(name="a", dims=[a_feature_dim, reduce_dim, batch_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[batch_dim, reduce_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[reduce_dim])

    assert result.dims == (batch_dim, a_feature_dim)
    assert torch.allclose(result.raw_tensor, torch.tensor([[5.0, 11.0, 17.0], [-10.0, -22.0, -34.0]]))


def test_dot_batched_matrix_matrix_product():
    a_raw = torch.tensor([[[1.0, 2.0], [-1.0, -2.0]], [[3.0, 4.0], [-3.0, -4.0]], [[5.0, 6.0], [-5.0, -6.0]]])
    b_raw = torch.tensor([[[1.0, 1.0], [2.0, 2.0]], [[3.0, 3.0], [4.0, 4.0]], [[5.0, 5.0], [6.0, 6.0]]])

    batch_dim = Dim(dimension=2)
    a_feature_dim = Dim(dimension=3)
    b_feature_dim = Dim(dimension=3)
    reduce_dim = Dim(dimension=2)

    # intentionally test strange batch axis
    a = Tensor(name="a", dims=[a_feature_dim, reduce_dim, batch_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[b_feature_dim, batch_dim, reduce_dim], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[reduce_dim])

    assert result.dims == (batch_dim, a_feature_dim, b_feature_dim)
    assert torch.allclose(result.raw_tensor, torch.zeros(size=(2, 3, 3)))  # values chosen such that everything cancels


def test_dot_batched_scale_matrix():
    a_raw = torch.tensor([2.0, 3.0])
    b_raw = torch.tensor([[[1.0, 2.0, 3.0], [-1.0, -2.0, -3.0]], [[2.0, 3.0, 4.0], [-2.0, -3.0, -4.0]]])

    batch_dim = Dim(dimension=2)
    b_feature_dim1 = Dim(dimension=2)
    b_feature_dim2 = Dim(dimension=3)

    a = Tensor(name="a", dims=[batch_dim], dtype="float32", raw_tensor=a_raw)
    b = Tensor(name="b", dims=[batch_dim, b_feature_dim1, b_feature_dim2], dtype="float32", raw_tensor=b_raw)

    result = rf.dot(a, b, reduce=[])

    assert result.dims == (batch_dim, b_feature_dim1, b_feature_dim2)
    assert torch.allclose(
        result.raw_tensor,
        torch.tensor([[[2.0, 4.0, 6.0], [-2.0, -4.0, -6.0]], [[6.0, 9.0, 12.0], [-6.0, -9.0, -12.0]]]),
    )


def test_dot_multiple_dims():
    a_raw = torch.rand(size=(2, 4, 6, 9, 5, 3, 8, 1))
    b_raw = torch.rand(size=(7, 2, 6, 8, 3, 1, 5, 4))

    reduce_dim_1 = Dim(dimension=3)
    reduce_dim_2 = Dim(dimension=6)
    reduce_dim_3 = Dim(dimension=1)

    common_dim_1 = Dim(dimension=2)
    common_dim_2 = Dim(dimension=8)
    common_dim_3 = Dim(dimension=5)

    a_unique_dim_1 = Dim(dimension=9)
    a_unique_dim_2 = Dim(dimension=4)

    b_unique_dim_1 = Dim(dimension=7)
    b_unique_dim_2 = Dim(dimension=4)

    a = Tensor(
        name="a",
        dims=[
            common_dim_1,
            a_unique_dim_2,
            reduce_dim_2,
            a_unique_dim_1,
            common_dim_3,
            reduce_dim_1,
            common_dim_2,
            reduce_dim_3,
        ],
        dtype="float32",
        raw_tensor=a_raw,
    )
    b = Tensor(
        name="b",
        dims=[
            b_unique_dim_1,
            common_dim_1,
            reduce_dim_2,
            common_dim_2,
            reduce_dim_1,
            reduce_dim_3,
            common_dim_3,
            b_unique_dim_2,
        ],
        dtype="float32",
        raw_tensor=b_raw,
    )

    result = rf.dot(a, b, reduce=[reduce_dim_1, reduce_dim_2, reduce_dim_3])

    # assumes common dims as sorted in a, unique dims as sorted in a / b respectively
    assert result.dims == (
        common_dim_1,
        common_dim_3,
        common_dim_2,
        a_unique_dim_2,
        a_unique_dim_1,
        b_unique_dim_1,
        b_unique_dim_2,
    )
    assert result.raw_tensor.shape == (2, 5, 8, 4, 9, 7, 4)



import sys
import _setup_test_env  # noqa
import unittest
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_greater, assert_true, assert_false
from pprint import pprint
from returnn.config import Config
from returnn.util.basic import PY3
from returnn.util import better_exchook

better_exchook.replace_traceback_format_tb()
if PY3:
    from io import StringIO
else:
    # noinspection PyUnresolvedReferences,PyCompatibility
    from StringIO import StringIO


def test_old_format():
    config = Config()
    config.load_file(
        StringIO(
            """
  # comment
  num_inputs 3
  hidden_type forward,lstm
  """
        )
    )

    assert_true(config.has("num_inputs"))
    assert_true(config.has("hidden_type"))
    assert_equal(config.int("num_inputs", -1), 3)
    assert_equal(config.value("hidden_type", "x"), "forward,lstm")
    assert_equal(config.value("hidden_type", "x", index=0), "forward")
    assert_equal(config.value("hidden_type", "x", index=1), "lstm")
    assert_equal(config.list("hidden_type", ["x"]), ["forward", "lstm"])

    assert_false(config.is_typed("num_inputs"))
    assert_false(config.is_typed("hidden_type"))


def test_json_format():
    config = Config()
    config.load_file(
        StringIO(
            """
  {
  // comment
  "num_inputs": 3,
  "hidden_type": ["forward", "lstm"]
  }
  """
        )
    )

    assert_true(config.has("num_inputs"))
    assert_true(config.has("hidden_type"))
    assert_equal(config.int("num_inputs", -1), 3)
    assert_equal(config.value("hidden_type", "x"), "forward,lstm")
    assert_equal(config.value("hidden_type", "x", index=0), "forward")
    assert_equal(config.value("hidden_type", "x", index=1), "lstm")
    assert_equal(config.list("hidden_type", ["x"]), ["forward", "lstm"])

    assert_true(config.is_typed("num_inputs"))
    assert_true(config.is_typed("hidden_type"))
    assert_is_instance(config.typed_value("num_inputs"), int)
    assert_is_instance(config.typed_value("hidden_type"), list)
    assert_equal(config.typed_value("hidden_type"), ["forward", "lstm"])


def test_py_config():
    config = Config()
    config.load_file(
        StringIO(
            """#!rnn.py
# comment
num_inputs = 3
hidden_type = ["forward", "lstm"]
  """
        )
    )

    assert_true(config.has("num_inputs"))
    assert_true(config.has("hidden_type"))
    assert_equal(config.int("num_inputs", -1), 3)
    assert_equal(config.value("hidden_type", "x"), "forward,lstm")
    assert_equal(config.value("hidden_type", "x", index=0), "forward")
    assert_equal(config.value("hidden_type", "x", index=1), "lstm")
    assert_equal(config.list("hidden_type", ["x"]), ["forward", "lstm"])

    assert_true(config.is_typed("num_inputs"))
    assert_true(config.is_typed("hidden_type"))
    assert_is_instance(config.typed_value("num_inputs"), int)
    assert_is_instance(config.typed_value("hidden_type"), list)
    assert_equal(config.typed_value("hidden_type"), ["forward", "lstm"])


def test_rnn_init_config_py_global_var():
    import returnn.__main__ as rnn
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".config", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """#!rnn.py
task = config.value("task", "train")

test_value = 42

def test_func():
  return task

    """
        )
        cfgfile.flush()
        rnn.init_config(command_line_options=[cfgfile.name, "--task", "search"])

    assert isinstance(rnn.config, Config)
    rnn.config.typed_dict.pop("__builtins__", None)  # not needed, too verbose for pprint
    pprint(rnn.config.dict)
    pprint(rnn.config.typed_dict)
    assert rnn.config.has("task")
    assert rnn.config.has("test_value")
    assert rnn.config.has("test_func")
    assert_equal(rnn.config.value("task", None), "search")
    assert rnn.config.is_typed("test_value")
    assert_equal(rnn.config.typed_value("test_value"), 42)
    assert rnn.config.is_typed("test_func")
    # So far it's fine.
    # Now something a bit strange.
    # Earlier, this failed, because the command-line overwrote this.
    assert rnn.config.is_typed("task")
    test_func = rnn.config.typed_dict["test_func"]
    assert callable(test_func)
    assert_equal(test_func(), "search")


def test_rnn_init_config_py_cmd_type():
    import returnn.__main__ as rnn
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".config", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """#!rnn.py
max_seq_length = {'bpe': 75}

def test_func():
  return max_seq_length

    """
        )
        cfgfile.flush()
        rnn.init_config(command_line_options=[cfgfile.name, "++max_seq_length", "0"])

    assert isinstance(rnn.config, Config)
    assert rnn.config.has("max_seq_length")
    assert rnn.config.has("test_func")
    assert rnn.config.is_typed("max_seq_length")
    assert rnn.config.is_typed("test_func")
    test_func = rnn.config.typed_dict["test_func"]
    assert callable(test_func)
    assert_equal(test_func(), 0)


def test_config_py_ext():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """
def test_func():
  return config.value("task", "train")
    """
        )
        cfgfile.flush()
        config = Config()
        config.load_file(cfgfile.name)  # should determine format by suffix ".py"

    config.typed_dict.pop("__builtins__", None)  # not needed, too verbose for pprint
    pprint(config.dict)
    pprint(config.typed_dict)
    assert config.has("test_func")
    assert config.is_typed("test_func")
    test_func = config.typed_dict["test_func"]
    assert callable(test_func)
    assert_equal(test_func(), "train")
    config.set("task", "search")
    assert_equal(test_func(), "search")


def test_config_py_old_returnn_imports():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """
# These are some common imports found in older config files.
from Pretrain import WrapEpochValue
from TFUtil import where_bc
import TFUtil
import returnn.TFUtil

import TFHorovod
hvd = TFHorovod.get_ctx(config=config)  # should return None if no Horovod context

from returnn.Util import describe_returnn_version
returnn_version = describe_returnn_version()

tf_version_tuple = returnn.TFUtil.tf_version_tuple()

    """
        )
        cfgfile.flush()
        config = Config()
        config.load_file(cfgfile.name)  # should determine format by suffix ".py"

    config.typed_dict.pop("__builtins__", None)  # not needed, too verbose for pprint
    pprint(config.typed_dict)

    import returnn.util.basic as util
    import returnn.tf.util.basic as tf_util

    assert config.typed_dict["where_bc"] is tf_util.where_bc
    assert config.typed_dict["TFUtil"].where_bc is tf_util.where_bc
    assert config.typed_dict["hvd"] is None
    assert config.typed_dict["tf_version_tuple"] == tf_util.tf_version_tuple()
    assert config.typed_dict["returnn_version"] == util.describe_returnn_version()
    assert config.typed_dict["returnn"].TFUtil is tf_util


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
Backends for the frontend API
"""

from __future__ import annotations
import typing
from typing import TYPE_CHECKING, Optional, Any, Union, TypeVar, Generic, Type, Sequence, Dict, Tuple
import numpy
import contextlib

from returnn.util.basic import NotSpecified

if TYPE_CHECKING:
    from returnn.tensor import Tensor, Dim
    from .types import RawTensorTypes as _RawTensorTypes

T = TypeVar("T")  # tf.Tensor, torch.Tensor or so
T2 = TypeVar("T2")


class Backend(Generic[T]):
    """
    Abstract base class for the backend, operating on tensor type T, i.e. :class:`Tensor[T]`.

    This class and instances do not have any state,
    and all functions are staticmethod (or classmethod).
    """

    # class attribs set by derived classes
    RawTensorType: Type[T]
    is_tensorflow: bool = False  # whether this framework uses TensorFlow

    # class private attribs
    _default_int_dtype: str = "int32"
    _default_float_dtype: str = "float32"

    def __init__(self):
        raise Exception("do not instantiate this class")

    # --- some generic helpers

    @classmethod
    def get_default_int_dtype(cls) -> str:
        """
        :return: default dtype for int
        """
        return cls._default_int_dtype

    @classmethod
    def get_default_float_dtype(cls) -> str:
        """
        :return: default dtype for float
        """
        return cls._default_float_dtype

    @classmethod
    def get_current_run_ctx(cls) -> RunCtx:
        """
        :return: current run context, see :class:`RunCtx`
        """
        pass  # TODO...

    # --- some functions which provide some reasonable default implementation

    # noinspection PyNestedDecorators
    @typing.overload
    @classmethod
    def compare(
        cls,
        a: Tensor,
        kind: str,
        b: Tensor,
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """compare with two tensors"""

    @classmethod
    def compare(
        cls,
        a: Union[Tensor, _RawTensorTypes],
        kind: str,
        b: Union[Tensor, _RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """
        :param a:
        :param kind: "equal"|"==", "less"|"<", "less_equal"|"<=", "greater"|">", "greater_equal"|">=", "not_equal"|"!="
        :param b:
        :param allow_broadcast_all_sources: if True, it is allowed that neither a nor b has all dims of the result.
            Not needed when out_dims is specified explicitly.
        :param dim_order: defines the order of the resulting dims. if None, it is automatically inferred from a and b.
            Not all the dims of a and b need to be specified here, and there could also be other dims in the dim_order.
        :return: element-wise comparison of a and b
        """
        from . import _utils as utils

        out, a, b = utils.bin_op_out_template(
            cls,
            a,
            b,
            name="compare",
            dtype="bool",
            allow_broadcast_all_sources=allow_broadcast_all_sources,
            dim_order=dim_order,
        )
        out.raw_tensor = cls.compare_raw(a.raw_tensor, kind, b.raw_tensor)
        return out

    # noinspection PyNestedDecorators
    @typing.overload
    @classmethod
    def combine(
        cls,
        a: Tensor,
        kind: str,
        b: Tensor,
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """combine with two tensors"""

    @classmethod
    def combine(
        cls,
        a: Union[Tensor, _RawTensorTypes],
        kind: str,
        b: Union[Tensor, _RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Union[Tensor, _RawTensorTypes]:
        """
        :param a:
        :param kind: "add"|"+", "sub"|"-", "mul"|"*", "truediv"|"/", "floordiv"|"//", "mod"|"%", "pow"|"**",
            "max"|"maximum", "min"|"minimum", "logical_and", "logical_or", "squared_difference"
        :param b:
        :param allow_broadcast_all_sources: if True, it is allowed that neither a nor b has all dims of the result.
            Not needed when out_dims is specified explicitly.
        :param dim_order: defines the order of the resulting dims. if None, it is automatically inferred from a and b.
            Not all the dims of a and b need to be specified here, and there could also be other dims in the dim_order.
        :return: element-wise combination of a and b
        """
        from . import _utils as utils

        if isinstance(b, (int, float, bool, numpy.number)):
            if b == 1 and kind in {"truediv", "/", "floordiv", "//", "pow", "**", "mul", "*"}:
                return a
            if b == -1 and kind in {"truediv", "/", "floordiv", "//", "pow", "**", "mul", "*"}:
                return -a
            if b == 0 and kind in {"add", "+", "sub", "-"}:
                return a
            if b and kind in {"logical_and", "logical_or"}:
                return a
        if isinstance(a, (int, float, bool, numpy.number)):
            if a == 1 and kind in {"pow", "**"}:
                return a
            if a == 1 and kind in {"mul", "*"}:
                return b
            if a == 0 and kind in {"add", "+", "sub", "-"}:
                return b
            if a and kind in {"logical_and", "logical_or"}:
                return b
        # Truediv checks for int/int division
        if kind in {"truediv", "/"}:
            if utils.is_int(cls, a) and utils.is_int(cls, b):
                raise ValueError(
                    "Dividing a Tensor of type int by an integer is disallowed. Please convert the Tensor to float."
                )
        out, a, b = utils.bin_op_out_template(
            cls,
            a,
            b,
            name="combine",
            dtype=a.dtype,
            allow_broadcast_all_sources=allow_broadcast_all_sources,
            dim_order=dim_order,
        )
        out.raw_tensor = cls.combine_raw(a.raw_tensor, kind, b.raw_tensor)
        return out

    @classmethod
    def mark_as_default_output(cls, tensor: Tensor, *, shape: Optional[Sequence[int]] = None) -> None:
        """
        Calls mark_as_output(tensor, "output", shape=shape).

        :param tensor:
        :param shape: see :func:`mark_as_output`
        """
        cls.mark_as_output(tensor, "output", shape=shape)

    # --- functions to override

    @staticmethod
    def get_dtype_name_raw(raw_tensor: T) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        raise NotImplementedError

    @staticmethod
    def get_ndim_raw(raw_tensor: T) -> int:
        """
        :return: ndim of raw tensor. assumes it is known
        """
        raise NotImplementedError

    @staticmethod
    def get_shape_raw(raw_tensor: T) -> T:
        """
        :return: shape of raw tensor
        """
        raise NotImplementedError

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: T) -> Tuple[Union[int, T]]:
        """
        :return: shape of raw tensor. assumes that ndim is known.
            In eager frameworks, all dims are int.
        """
        raise NotImplementedError

    @staticmethod
    def get_known_shape_raw(raw_tensor: T) -> Tuple[Optional[int]]:
        """
        :return: shape of raw tensor, int for static known, None otherwise. assumes that ndim is known.
            This will not create any ops.
            In eager frameworks, all dims are known.
        """
        raise NotImplementedError

    @staticmethod
    def set_known_shape_raw(raw_tensor: T, shape: Tuple[Optional[int]]) -> None:
        """
        Sets the known shape of the raw tensor.
        This is only supported in graph-based frameworks,
        and just performs a check in eager frameworks.
        """
        # Default implementation for eager-based frameworks.
        assert all(dim is not None for dim in shape)
        rf = get_backend_by_raw_tensor_type(type(raw_tensor))
        existing_shape = rf.get_known_shape_raw(raw_tensor)
        assert all(dim is not None for dim in existing_shape)
        assert shape == existing_shape

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, T]], T], value: Union[Any, T]) -> T:
        """
        :param shape: shape
        :param value: scalar value to fill
        :return: raw tensor filled with value everywhere
        """
        raise NotImplementedError

    @staticmethod
    def compare_raw(a: T, kind: str, b: T) -> T:
        """
        :param a:
        :param kind: "equal"|"==", "less"|"<", "less_equal"|"<=", "greater"|">", "greater_equal"|">=", "not_equal"|"!="
        :param b:
        :return: a `kind` b
        """
        raise NotImplementedError

    @staticmethod
    def combine_raw(a: T, kind: str, b: T) -> T:
        """
        :param a:
        :param kind: "add"|"+", "sub"|"-", "mul"|"*", "truediv"|"/", "floordiv"|"//", "mod"|"%", "pow"|"**",
            "max"|"maximum", "min"|"minimum", "logical_and", "logical_or", "squared_difference"
        :param b:
        :return: a `kind` b
        """
        raise NotImplementedError

    @staticmethod
    def reshape_raw(raw_tensor: T, shape: Union[Sequence[Union[int, T]], T]) -> T:
        """
        :param raw_tensor: raw tensor
        :param shape: new shape
        :return: reshaped raw tensor
        """
        raise NotImplementedError

    @classmethod
    def squeeze_raw(cls, raw_tensor: T, axes: Sequence[int]) -> T:
        """
        :param raw_tensor: raw tensor
        :param axes: axes to squeeze
        :return: squeezed raw tensor
        """
        # Default implementation using reshape_raw.
        known_shape = cls.get_known_shape_raw(raw_tensor)
        assert all([known_shape[axis] == 1 for axis in axes])
        new_shape = [dim for a, dim in enumerate(cls.get_shape_tuple_raw(raw_tensor)) if a not in axes]
        return cls.reshape_raw(raw_tensor, new_shape)

    @staticmethod
    def transpose_raw(raw_tensor: T, perm: Sequence[int]) -> T:
        """
        :param raw_tensor: raw tensor
        :param perm: permutation
        :return: transposed raw tensor
        """
        raise NotImplementedError

    @staticmethod
    def expand_dims_raw(raw_tensor: T, axis: int) -> T:
        """
        :param raw_tensor:
        :param axis:
        :return: raw tensor with new axis
        """
        raise NotImplementedError

    @staticmethod
    def expand_raw(raw_tensor: T, axis: int, dim: Union[int, T]) -> T:
        """
        :param raw_tensor:
        :param axis: shape[axis] must be 1
        :param dim: the new dim for shape[axis]
        :return: shape[axis] expands to dim.
            in PyTorch or other frameworks which support custom strides,
            this is an efficient view and not a copy.
        """
        raise NotImplementedError

    @staticmethod
    def sequence_mask_raw(lengths: T, *, batch_major: bool = True) -> T:
        """
        Like tf.sequence_mask().

        :param lengths: shape (batch,)
        :param batch_major:
        :return: tensor mask of shape (batch,maxlen) if batch_major else (maxlen,batch) of type bool
        """
        raise NotImplementedError

    @staticmethod
    @contextlib.contextmanager
    def name_scope_raw(name: str) -> Any:
        """
        Default implementation for eager-based frameworks:
        Do nothing, tensors do not have a name.

        :param name:
        :return: context manager
        """
        # Default implementation for eager-based frameworks
        pass  # nothing to do

    @staticmethod
    @contextlib.contextmanager
    def control_dependencies_raw(dependencies: Sequence[Any]) -> Any:
        """
        Default implementation for eager-based frameworks:
        Do nothing, we expect that the dependencies are already executed.

        :param dependencies: raw tensors or ops
        :return: context manager
        """
        # Default implementation for eager-based frameworks
        yield

    @staticmethod
    def identity_with_control_dependencies_raw(raw_tensor: T, dependencies: Sequence[Any]) -> T:
        """
        Default implementation for eager-based frameworks:
        Do nothing, we expect that the dependencies are already executed.

        :param raw_tensor: raw tensor
        :param dependencies: raw tensors or ops
        :return: raw tensor
        """
        # Default implementation for eager-based frameworks
        return raw_tensor

    @staticmethod
    def create_placeholder(tensor: Tensor) -> T:
        """
        :return: tf.placeholder in TF

        This is really only for TensorFlow for the deprecated option auto_create_placeholders
        and should not be used in other backends,
        even in graph-based backends.
        Rather, the logic to create placeholders should be done elsewhere.
        """
        raise Exception("create_placeholder not supported by backend")

    @staticmethod
    def runtime_sanity_checks(tensor: Tensor) -> Any:
        """
        Checks whether the tensor.raw_tensor is consistent with the tensor metadata.

        In graph-based frameworks (TF graph), we return some operation here.
        In eager frameworks, we would not return anything but instead directly perform the checks.
        """
        # By default, we do not do any checks. This is optional for the backend.
        pass

    @staticmethod
    def is_valid_in_current_graph(tensor: Tensor) -> bool:
        """
        :return: whether the raw tensor is valid in the current graph.
            In eager-mode frameworks, this is always true -- there is no graph.
        """
        return True

    @staticmethod
    def format_graph_output(raw_tensor: T, *, max_depth: Optional[int] = None) -> str:
        """
        :return: the computation graph leading to this tensor formatted.
            In eager-mode frameworks, this is not supported and returns None.
        """
        return "<no-graph>"

    @staticmethod
    def mark_as_loss(
        loss: Tensor,
        name: str,
        *,
        scale: Optional[float] = 1.0,
        as_error: bool = False,
        use_normalized_loss: bool = False,
        use_flatten_frames: bool = True,
        custom_inv_norm_factor: Optional[Tensor] = None,
    ) -> None:
        """
        Mark the given loss tensor as a loss.
        This has the effect that it is specially handled by RETURNN.
        Specifically, the optimizer can use it in training,
        and it is used for reporting per batch or per epoch,
        and for learning rate scheduling.

        This currently uses :class:`AsIsLoss` in RETURNN
        but this is an implementation detail and might change.

        :param loss:
        :param name: name of the loss. this name is used for reporting by RETURNN, and also for LR scheduling.
        :param scale: scale the loss by this factor for the training optimizer
          (but not for any reporting). setting to 0.0 has the effect that this loss is not used by the optimizer.
        :param as_error: if True, this loss is reported as an error instead of a loss,
          and not used by the training optimizer.
          This is by convention sth like the frame-error or edit-distance, and usually not differentiable anyway.
        :param bool use_flatten_frames: If True, will use :func:`returnn.tf.util.basic.flatten_with_seq_len_mask`,
          i.e. a "packed" sequence with the padded frames removed, and accumulates over that.
          This can be more efficient, also because it will further optimize incoming computations
          and e.g. skip softmax computations right before on the padded frames.
          This can also avoid issues with inf/nan in some cases.
          If False, it will mask the loss to 0 in the padded frames and accumulate over that.
          Typically, setting this to True (default) is both more efficient and better.
        :param bool use_normalized_loss: the loss used in optimization will be normalized.
          E.g. if the overall normalization is sum(loss)/sum(num_frames), this is also what the optimizer will use,
          otherwise the optimizer will just use sum(loss).
        :param custom_inv_norm_factor:
          The standard norm factor is 1/sum(target_seq_len) if the target has a time-axis,
          or 1/sum(output_seq_len) if there is no target and the output has a time-axis,
          or 1 otherwise. (See :func:`Loss.init` for details.)
          This is used for proper normalization of accumulated loss/error per epoch
          and also proper normalization per batch for reporting,
          no matter if use_normalized_loss is True or False.
          If you want to change this norm factor, you can set this.
          Basically, for all reporting, it uses sum(loss) * sum(custom_inv_norm_factor).
        """
        raise NotImplementedError

    @staticmethod
    def mark_as_output(tensor: Tensor, name: str, *, shape: Optional[Sequence[int]] = None) -> None:
        """
        Mark this as an output.
        This has the effect that RETURNN will in any case construct the corresponding layer.
        Also see :func:`mark_as_default_output`.

        This is intended mostly for forwarding, or exporting the model (TF graph, TFLite, ONNX, etc).
        You must specify a shape to have the output shape (order of dims) well-defined
        (if not specified, we check if some defaults are possible, like BTF, or BF).

        :param tensor:
        :param name:
        :param shape: this specifies the order of the dims of the output, such that it is well-defined
            for some external application.
            If not specified, we try to infer BTF or BF as default, if that works, otherwise it will be an error.
        """
        raise NotImplementedError

    @staticmethod
    def convert_to_tensor(value: Union[Tensor, T, _RawTensorTypes]) -> Tensor[T]:
        """
        :param value: tensor, or scalar raw tensor or some other scalar value
        :return: tensor
        """
        raise NotImplementedError

    @staticmethod
    def dot(a: Tensor[T], b: Tensor[T], *, reduce: Union[Dim, Sequence[Dim]]) -> Tensor[T]:
        """
        This performs a dot-product of two sources a and b.
        The underlying operation is a batched matmul (shared..., I, J) * (shared..., J, K) -> (shared..., I, K).
        The inputs a and b are transformed internally into the required shapes in the following way:
        The axis J is specified via the Dim given as 'reduce'. If multiple reduce Dims are given the corresponding axes
        are merged into one before the matmul via a reshape. All other matching Dims in a and b will be treated as
        batch dimensions ('shared...'). Dims unique to a and b define the axes I and K, respectively. (Multiple or no
        unique axes in a and b are supported too.)

        Depending on which Dims exist in a, b and reduce this dot operation can be used to compute scaling, scalar
        product, outer product, matrix-vector multiplication, matrix-matrix multiplication etc. (all possibly batched).

        :param a:
        :param b:
        :param reduce: Dims over which to perform the product, have to be present in both a and b
        :return: result of dot product, Dim order: common axes as sorted in a, unique axes of a (in order),
            unique axes of b (in order)
        """
        raise NotImplementedError

    @staticmethod
    def range_over_dim(dim: Dim) -> Tensor[T]:
        """
        :param dim:
        :return: tensor with shape [dim]
        """
        raise NotImplementedError

    @staticmethod
    def reduce(
        source: Tensor[T],
        *,
        mode: str,
        axis: Union[Dim, Sequence[Dim]],
        use_time_mask: bool = NotSpecified,
    ) -> Tensor[T]:
        """
        Reduce the tensor along the given axis

        :param source:
        :param mode: "sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"
        :param axis:
        :param use_time_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
        :return: tensor with axis removed
        """
        raise NotImplementedError


# We use a global instance, and we modify __class__ inplace,
# such that any reference to this can be updated.
# This is exposed to the user as `returnn.frontend`.
# The __class__ assignment is done in `select_engine`.
# Use object.__new__ because we disallow creating instances of Frontend.
global_backend = object.__new__(Backend)

_dispatch_table = {}  # type: Dict[Type, Type[Backend]]


def select_backend_returnn_layers_tf():
    """
    Selects the RETURNN layers backend (based on TF).
    """
    import tensorflow as tf

    backend = get_backend_by_raw_tensor_type(tf.Tensor)  # side-effect: register it
    global_backend.__class__ = backend

    # TODO returnn layer type, register_frontend_by_tensor_type for that
    #   global_frontend.__class__ = ReturnnLayersFrontend


def select_backend_torch():
    """
    Selects the PyTorch (low-level) backend.
    """
    import torch

    backend = get_backend_by_raw_tensor_type(torch.Tensor)  # side-effect: register it
    global_backend.__class__ = backend


def get_backend_by_tensor(tensor: Tensor, *, fallback: Optional[T2] = None) -> Union[Type[Backend[T]], T2]:
    """
    :param tensor:
    :param fallback:
    """
    if fallback and tensor.raw_tensor is None:
        return fallback
    assert tensor.raw_tensor is not None
    return get_backend_by_raw_tensor_type(type(tensor.raw_tensor))


def get_backend_by_raw_tensor_type(tensor_type: Type[T]) -> Union[Type[Backend[T]]]:
    """
    :param tensor_type:
    """
    if tensor_type not in _dispatch_table:
        # We don't register all possible subclasses in the dispatch table.
        # Check through the MRO.
        for type_ in tensor_type.__mro__:
            if type_ in _dispatch_table:
                # Also register it for faster future lookups.
                _dispatch_table[tensor_type] = _dispatch_table[type_]
                return _dispatch_table[type_]
    if tensor_type not in _dispatch_table:
        # It would be registered if there was any select_engine or select_backend_* call.
        # However, some code might not have done that, so for the common cases,
        # we do it here.
        if tensor_type.__module__.split(".")[0] == "tensorflow":
            from returnn.tf.frontend_low_level import TFBackend

            backend_type = TFBackend
            tensor_types = _get_tensor_types_tf()
        elif tensor_type.__module__.split(".")[0] == "torch":
            from returnn.torch.frontend import TorchBackend

            backend_type = TorchBackend
            tensor_types = _get_tensor_types_torch()
        else:
            raise Exception(f"unknown tensor type {tensor_type}")
        assert any(issubclass(tensor_type, type_) for type_ in tensor_types)
        for type_ in tensor_types:
            register_backend_by_tensor_type(type_, backend_type)
        return backend_type
    return _dispatch_table[tensor_type]


def register_backend_by_tensor_type(tensor_type: Type[T], backend: Type[Backend[T]]):
    """
    :param tensor_type:
    :param backend:
    """
    _dispatch_table[tensor_type] = backend


def _get_tensor_types_tf():
    """
    :return: tuple of relevant tensor types in TF.
        Note that it is not so important to cover all, as we also check issubclass as a fallback.
    """
    import tensorflow as tf

    ls = [tf.Tensor, tf.Variable]
    return tuple(ls)


def _get_tensor_types_torch():
    """
    :return: tuple of relevant tensor types in PyTorch.
        Note that it is not so important to cover all, as we also check issubclass as a fallback.
    """
    import torch

    ls = [torch.Tensor, torch.nn.Parameter]
    return tuple(ls)


class RunCtx:
    """
    We can either be in param-init stage,
    or in the main training loop,
    or forwarding loop.

    In training, we expect that some loss is being defined via mark_as_loss().
    In forwarding, we expect that some output is being defined via mark_as_output().
    """

    # TODO ...



"""
Tensor binary operations mixin.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .tensor import Tensor  # just for type hints; otherwise use _t.Tensor
    import returnn.frontend.types as _rf_types

from ._tensor_mixin_base import _TensorMixinBase


class _TensorOpOverloadsMixin(_TensorMixinBase):

    # --- comparisons

    def __eq__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.compare(self, "==", other)

    def __ne__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.compare(self, "!=", other)

    def __lt__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.compare(self, "<", other)

    def __le__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.compare(self, "<=", other)

    def __gt__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.compare(self, ">", other)

    def __ge__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.compare(self, ">=", other)

    # --- math binary and unary ops

    def __add__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "+", other)

    def __radd__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "+", self)

    def __sub__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "-", other)

    def __rsub__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "-", self)

    def __mul__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "*", other)

    def __rmul__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "*", self)

    def __truediv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "/", other)

    def __rtruediv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "/", self)

    def __floordiv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "//", other)

    def __rfloordiv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "//", self)

    def __mod__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "%", other)

    def __rmod__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "%", self)

    def __pow__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "**", other)

    def __rpow__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "**", self)

    def __neg__(self):  # -x
        if True:  # avoid warning: abstract base class...
            raise NotImplementedError  # TODO

    def __invert__(self):  # ~x
        if True:
            raise NotImplementedError  # TODO

    def __abs__(self):
        if True:
            raise NotImplementedError  # TODO

    def __ceil__(self):
        if True:
            raise NotImplementedError  # TODO

    def __floor__(self):
        if True:
            raise NotImplementedError  # TODO

    def __and__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "logical_and", other)

    def __rand__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "logical_and", self)

    def __or__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(self, "logical_or", other)

    def __ror__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return self._raw_backend.combine(other, "logical_or", self)



#!/usr/bin/env python

"""
Dumps the network topology as JSON on stdout.
"""

from __future__ import annotations

import sys
import argparse
import json
import typing

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
from returnn.pretrain import pretrain_from_config
from returnn.config import network_json_from_config


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_filename, command_line_options):
    """
    :param str config_filename:
    :param list[str] command_line_options:
    """
    rnn.init_better_exchook()
    rnn.init_config(config_filename, command_line_options)
    global config
    config = rnn.config
    config.set("log", [])
    rnn.init_log()
    print("RETURNN dump-dataset starting up.", file=log.v3)
    rnn.init_config_json_network()


def main(argv):
    """
    Main entry.
    """
    arg_parser = argparse.ArgumentParser(description="Dump network as JSON.")
    arg_parser.add_argument("returnn_config_file")
    arg_parser.add_argument("--epoch", default=1, type=int)
    arg_parser.add_argument("--out", default="/dev/stdout")
    args = arg_parser.parse_args(argv[1:])
    init(config_filename=args.returnn_config_file, command_line_options=[])

    pretrain = pretrain_from_config(config)
    if pretrain:
        json_data = pretrain.get_network_json_for_epoch(args.epoch)
    else:
        json_data = network_json_from_config(config)

    f = open(args.out, "w")
    print(json.dumps(json_data, indent=2, sort_keys=True), file=f)
    f.close()

    rnn.finalize()


if __name__ == "__main__":
    main(sys.argv)



"""
High-level backend for RETURNN layers
"""

from __future__ import annotations
from typing import Union, Sequence

from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim

# noinspection PyProtectedMember
from returnn.frontend._backend import Backend


# Ignore this warning until we really expect that we implemented everything.
# noinspection PyAbstractClass
class ReturnnLayersBackend(Backend):
    """
    RETURNN layers backend (using TF), where raw_tensor represents a RETURNN layer
    """

    is_tensorflow = True

    @staticmethod
    def reduce(
        source: Tensor, *, mode: str, axis: Union[Dim, Sequence[Dim]], use_time_mask: bool = NotSpecified
    ) -> Tensor:
        """Reduce"""
        raise NotImplementedError  # TODO



"""
Utilities for dimension tags, dimensions, axes.
"""


from __future__ import annotations
from typing import TypeVar
from returnn.tensor import Tensor, Dim
from ._backend import get_backend_by_tensor, global_backend

T = TypeVar("T")


def range_over_dim(dim: Dim) -> Tensor[T]:
    """
    :param dim:
    :return: tensor with shape [dim]
    """
    if dim.dyn_size_ext:
        rf = get_backend_by_tensor(dim.dyn_size_ext, fallback=global_backend)
    else:
        rf = global_backend
    return rf.range_over_dim(dim)
